# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['TestEnvironmentSchema.test_basic_valid_config_on_run_config_schema[readonly_in_memory_instance_in_process_env] 1'] = {
    'runConfigSchemaOrError': {
        'isRunConfigValid': {
            '__typename': 'PipelineConfigValidationValid',
            'pipelineName': 'csv_hello_world'
        }
    }
}

snapshots['TestEnvironmentSchema.test_basic_valid_config_on_run_config_schema[readonly_in_memory_instance_managed_grpc_env] 1'] = {
    'runConfigSchemaOrError': {
        'isRunConfigValid': {
            '__typename': 'PipelineConfigValidationValid',
            'pipelineName': 'csv_hello_world'
        }
    }
}

snapshots['TestEnvironmentSchema.test_basic_valid_config_on_run_config_schema[readonly_in_memory_instance_multi_location] 1'] = {
    'runConfigSchemaOrError': {
        'isRunConfigValid': {
            '__typename': 'PipelineConfigValidationValid',
            'pipelineName': 'csv_hello_world'
        }
    }
}

snapshots['TestEnvironmentSchema.test_basic_valid_config_on_run_config_schema[readonly_sqlite_instance_deployed_grpc_env] 1'] = {
    'runConfigSchemaOrError': {
        'isRunConfigValid': {
            '__typename': 'PipelineConfigValidationValid',
            'pipelineName': 'csv_hello_world'
        }
    }
}

snapshots['TestEnvironmentSchema.test_basic_valid_config_on_run_config_schema[readonly_sqlite_instance_in_process_env] 1'] = {
    'runConfigSchemaOrError': {
        'isRunConfigValid': {
            '__typename': 'PipelineConfigValidationValid',
            'pipelineName': 'csv_hello_world'
        }
    }
}

snapshots['TestEnvironmentSchema.test_basic_valid_config_on_run_config_schema[readonly_sqlite_instance_managed_grpc_env] 1'] = {
    'runConfigSchemaOrError': {
        'isRunConfigValid': {
            '__typename': 'PipelineConfigValidationValid',
            'pipelineName': 'csv_hello_world'
        }
    }
}

snapshots['TestEnvironmentSchema.test_basic_valid_config_on_run_config_schema[readonly_sqlite_instance_multi_location] 1'] = {
    'runConfigSchemaOrError': {
        'isRunConfigValid': {
            '__typename': 'PipelineConfigValidationValid',
            'pipelineName': 'csv_hello_world'
        }
    }
}



from dagster import check
from dagster.core.host_representation import (
    ExternalRepository,
    ExternalRepositoryOrigin,
    GrpcServerRepositoryLocationHandle,
    ManagedGrpcPythonEnvRepositoryLocationHandle,
    RepositoryLocation,
)
from dagster.core.host_representation.grpc_server_state_subscriber import (
    LocationStateChangeEventType,
)
from dagster.utils.error import SerializableErrorInfo
from dagster_graphql import dauphin
from dagster_graphql.implementation.fetch_solids import get_solid, get_solids
from dagster_graphql.schema.errors import DauphinPythonError

DauphinLocationStateChangeEventType = dauphin.Enum.from_enum(LocationStateChangeEventType)


class DauphinRepository(dauphin.ObjectType):
    class Meta:
        name = "Repository"

    def __init__(self, repository, repository_location):
        self._repository = check.inst_param(repository, "repository", ExternalRepository)
        self._repository_location = check.inst_param(
            repository_location, "repository_location", RepositoryLocation
        )
        super(DauphinRepository, self).__init__(name=repository.name)

    id = dauphin.NonNull(dauphin.ID)
    name = dauphin.NonNull(dauphin.String)
    location = dauphin.NonNull("RepositoryLocation")
    pipelines = dauphin.non_null_list("Pipeline")
    usedSolids = dauphin.Field(dauphin.non_null_list("UsedSolid"))
    usedSolid = dauphin.Field("UsedSolid", name=dauphin.NonNull(dauphin.String))
    origin = dauphin.NonNull("RepositoryOrigin")
    partitionSets = dauphin.non_null_list("PartitionSet")
    schedules = dauphin.non_null_list("Schedule")
    sensors = dauphin.non_null_list("Sensor")

    def resolve_id(self, _graphene_info):
        return self._repository.get_external_origin_id()

    def resolve_origin(self, graphene_info):
        origin = self._repository.get_external_origin()
        return graphene_info.schema.type_named("RepositoryOrigin")(origin)

    def resolve_location(self, graphene_info):
        return graphene_info.schema.type_named("RepositoryLocation")(self._repository_location)

    def resolve_schedules(self, graphene_info):

        schedules = self._repository.get_external_schedules()

        return sorted(
            [
                graphene_info.schema.type_named("Schedule")(graphene_info, schedule)
                for schedule in schedules
            ],
            key=lambda schedule: schedule.name,
        )

    def resolve_sensors(self, graphene_info):
        sensors = self._repository.get_external_sensors()
        return sorted(
            [
                graphene_info.schema.type_named("Sensor")(graphene_info, sensor)
                for sensor in sensors
            ],
            key=lambda sensor: sensor.name,
        )

    def resolve_pipelines(self, graphene_info):
        return sorted(
            [
                graphene_info.schema.type_named("Pipeline")(pipeline)
                for pipeline in self._repository.get_all_external_pipelines()
            ],
            key=lambda pipeline: pipeline.name,
        )

    def resolve_usedSolid(self, _graphene_info, name):
        return get_solid(self._repository, name)

    def resolve_usedSolids(self, _graphene_info):
        return get_solids(self._repository)

    def resolve_partitionSets(self, graphene_info):
        return (
            graphene_info.schema.type_named("PartitionSet")(self._repository.handle, partition_set)
            for partition_set in self._repository.get_external_partition_sets()
        )


class DauphinRepositoryOrigin(dauphin.ObjectType):
    class Meta:
        name = "RepositoryOrigin"

    repository_location_name = dauphin.NonNull(dauphin.String)
    repository_name = dauphin.NonNull(dauphin.String)
    repository_location_metadata = dauphin.non_null_list("RepositoryMetadata")

    def __init__(self, origin):
        self._origin = check.inst_param(origin, "origin", ExternalRepositoryOrigin)

    def resolve_repository_location_name(self, _graphene_info):
        return self._origin.repository_location_origin.location_name

    def resolve_repository_name(self, _graphene_info):
        return self._origin.repository_name

    def resolve_repository_location_metadata(self, graphene_info):
        metadata = self._origin.repository_location_origin.get_display_metadata()
        return [
            graphene_info.schema.type_named("RepositoryMetadata")(key=key, value=value)
            for key, value in metadata.items()
            if value is not None
        ]


class DauphinRepositoryMetadata(dauphin.ObjectType):
    class Meta:
        name = "RepositoryMetadata"

    key = dauphin.NonNull(dauphin.String)
    value = dauphin.NonNull(dauphin.String)


class DauphinRepositoryLocationOrLoadFailure(dauphin.Union):
    class Meta:
        name = "RepositoryLocationOrLoadFailure"
        types = ("RepositoryLocation", "RepositoryLocationLoadFailure")


class DauphinRepositoryLocation(dauphin.ObjectType):
    class Meta:
        name = "RepositoryLocation"

    id = dauphin.NonNull(dauphin.ID)
    name = dauphin.NonNull(dauphin.String)
    is_reload_supported = dauphin.NonNull(dauphin.Boolean)
    environment_path = dauphin.String()
    repositories = dauphin.non_null_list("Repository")
    server_id = dauphin.String()

    def __init__(self, location):
        self._location = check.inst_param(location, "location", RepositoryLocation)
        environment_path = (
            location.location_handle.executable_path
            if isinstance(location.location_handle, ManagedGrpcPythonEnvRepositoryLocationHandle)
            else None
        )

        server_id = (
            location.location_handle.server_id
            if isinstance(location.location_handle, GrpcServerRepositoryLocationHandle)
            else None
        )

        check.invariant(location.name is not None)

        super(DauphinRepositoryLocation, self).__init__(
            name=location.name,
            environment_path=environment_path,
            is_reload_supported=location.is_reload_supported,
            server_id=server_id,
        )

    def resolve_id(self, _):
        return self.name

    def resolve_repositories(self, graphene_info):
        return [
            graphene_info.schema.type_named("Repository")(repository, self._location)
            for repository in self._location.get_repositories().values()
        ]


class DauphinRepositoryLocationLoadFailure(dauphin.ObjectType):
    class Meta:
        name = "RepositoryLocationLoadFailure"

    id = dauphin.NonNull(dauphin.ID)
    name = dauphin.NonNull(dauphin.String)
    error = dauphin.NonNull("PythonError")

    def __init__(self, name, error):
        check.str_param(name, "name")
        check.inst_param(error, "error", SerializableErrorInfo)
        super(DauphinRepositoryLocationLoadFailure, self).__init__(
            name=name, error=DauphinPythonError(error)
        )

    def resolve_id(self, _):
        return self.name


class DauphinRepositoryConnection(dauphin.ObjectType):
    class Meta:
        name = "RepositoryConnection"

    nodes = dauphin.non_null_list("Repository")


class DauphinRepositoryLocationConnection(dauphin.ObjectType):
    class Meta:
        name = "RepositoryLocationConnection"

    nodes = dauphin.non_null_list("RepositoryLocationOrLoadFailure")


class DauphinLocationStateChangeSubscription(dauphin.ObjectType):
    class Meta(object):
        name = "LocationStateChangeSubscription"

    event = dauphin.Field(dauphin.NonNull("LocationStateChangeEvent"))


class DauphinLocationStateChangeEvent(dauphin.ObjectType):
    class Meta(object):
        name = "LocationStateChangeEvent"

    event_type = dauphin.NonNull("LocationStateChangeEventType")
    message = dauphin.NonNull(dauphin.String)
    location_name = dauphin.NonNull(dauphin.String)
    server_id = dauphin.Field(dauphin.String)


def get_location_state_change_observable(graphene_info):
    context = graphene_info.context
    return context.location_state_events.map(
        lambda event: graphene_info.schema.type_named("LocationStateChangeSubscription")(
            event=graphene_info.schema.type_named("LocationStateChangeEvent")(
                event_type=event.event_type,
                location_name=event.location_name,
                message=event.message,
                server_id=event.server_id,
            ),
        )
    )



import pendulum
import yaml
from dagster import check
from dagster.core.definitions.job import JobType, RunRequest
from dagster.core.host_representation import (
    ExternalScheduleExecutionData,
    ExternalScheduleExecutionErrorData,
)
from dagster.core.scheduler.job import (
    JobState,
    JobStatus,
    JobTick,
    JobTickStatus,
    ScheduleJobData,
    SensorJobData,
)
from dagster.core.storage.pipeline_run import PipelineRunsFilter
from dagster.core.storage.tags import TagType, get_tag_type
from dagster_graphql import dauphin
from dagster_graphql.implementation.fetch_schedules import get_schedule_next_tick
from dagster_graphql.implementation.fetch_sensors import get_sensor_next_tick


class DauphinJobTick(dauphin.ObjectType):
    class Meta:
        name = "JobTick"

    id = dauphin.NonNull(dauphin.ID)
    status = dauphin.NonNull("JobTickStatus")
    timestamp = dauphin.NonNull(dauphin.Float)
    runIds = dauphin.non_null_list(dauphin.String)
    error = dauphin.Field("PythonError")
    skipReason = dauphin.String()

    runs = dauphin.non_null_list("PipelineRun")

    def __init__(self, _, job_tick):
        self._job_tick = check.inst_param(job_tick, "job_tick", JobTick)

        super(DauphinJobTick, self).__init__(
            status=job_tick.status,
            timestamp=job_tick.timestamp,
            runIds=job_tick.run_ids,
            error=job_tick.error,
            skipReason=job_tick.skip_reason,
        )

    def resolve_id(self, _):
        return "%s:%s" % (self._job_tick.job_origin_id, self._job_tick.timestamp)

    def resolve_runs(self, graphene_info):
        instance = graphene_info.context.instance
        return [
            graphene_info.schema.type_named("PipelineRun")(instance.get_run_by_id(run_id))
            for run_id in self._job_tick.run_ids
            if instance.has_run(run_id)
        ]


class DauphinFutureJobTick(dauphin.ObjectType):
    class Meta(object):
        name = "FutureJobTick"

    timestamp = dauphin.NonNull(dauphin.Float)
    evaluationResult = dauphin.Field("TickEvaluation")

    def __init__(self, job_state, timestamp):
        self._job_state = check.inst_param(job_state, "job_state", JobState)
        self._timestamp = timestamp
        super(DauphinFutureJobTick, self).__init__(
            timestamp=check.float_param(timestamp, "timestamp"),
        )

    def resolve_evaluationResult(self, graphene_info):
        if self._job_state.status != JobStatus.RUNNING:
            return None

        if self._job_state.job_type != JobType.SCHEDULE:
            return None

        repository_origin = self._job_state.origin.external_repository_origin
        if not graphene_info.context.has_repository_location(
            repository_origin.repository_location_origin.location_name
        ):
            return None

        repository_location = graphene_info.context.get_repository_location(
            repository_origin.repository_location_origin.location_name
        )
        if not repository_location.has_repository(repository_origin.repository_name):
            return None

        repository = repository_location.get_repository(repository_origin.repository_name)
        external_schedule = repository.get_external_job(self._job_state.name)
        timezone_str = external_schedule.execution_timezone
        if not timezone_str:
            timezone_str = pendulum.now().timezone.name

        next_tick_datetime = next(external_schedule.execution_time_iterator(self._timestamp))
        schedule_time = pendulum.instance(next_tick_datetime).in_tz(timezone_str)
        schedule_data = repository_location.get_external_schedule_execution_data(
            instance=graphene_info.context.instance,
            repository_handle=repository.handle,
            schedule_name=external_schedule.name,
            scheduled_execution_time=schedule_time,
        )
        return graphene_info.schema.type_named("TickEvaluation")(schedule_data)


class DauphinTickEvaluation(dauphin.ObjectType):
    class Meta(object):
        name = "TickEvaluation"

    runRequests = dauphin.List("RunRequest")
    skipReason = dauphin.String()
    error = dauphin.Field("PythonError")

    def __init__(self, schedule_data):
        check.inst_param(
            schedule_data,
            "schedule_data",
            (ExternalScheduleExecutionData, ExternalScheduleExecutionErrorData),
        )
        error = (
            schedule_data.error
            if isinstance(schedule_data, ExternalScheduleExecutionErrorData)
            else None
        )
        skip_reason = (
            schedule_data.skip_message
            if isinstance(schedule_data, ExternalScheduleExecutionData)
            else None
        )
        self._run_requests = (
            schedule_data.run_requests
            if isinstance(schedule_data, ExternalScheduleExecutionData)
            else None
        )
        super(DauphinTickEvaluation, self).__init__(skipReason=skip_reason, error=error)

    def resolve_runRequests(self, graphene_info):
        if not self._run_requests:
            return self._run_requests

        return [
            graphene_info.schema.type_named("RunRequest")(run_request)
            for run_request in self._run_requests
        ]


class DauphinRunRequest(dauphin.ObjectType):
    class Meta(object):
        name = "RunRequest"

    runKey = dauphin.String()
    tags = dauphin.non_null_list("PipelineTag")
    runConfigYaml = dauphin.NonNull(dauphin.String)

    def __init__(self, run_request):
        super(DauphinRunRequest, self).__init__(runKey=run_request.run_key)
        self._run_request = check.inst_param(run_request, "run_request", RunRequest)

    def resolve_tags(self, graphene_info):
        return [
            graphene_info.schema.type_named("PipelineTag")(key=key, value=value)
            for key, value in self._run_request.tags.items()
            if get_tag_type(key) != TagType.HIDDEN
        ]

    def resolve_runConfigYaml(self, _graphene_info):
        return yaml.dump(self._run_request.run_config, default_flow_style=False, allow_unicode=True)


class DauphinFutureJobTicks(dauphin.ObjectType):
    class Meta(object):
        name = "FutureJobTicks"

    results = dauphin.non_null_list("FutureJobTick")
    cursor = dauphin.NonNull(dauphin.Float)


class DauphinJobState(dauphin.ObjectType):
    class Meta:
        name = "JobState"

    id = dauphin.NonNull(dauphin.ID)
    name = dauphin.NonNull(dauphin.String)
    jobType = dauphin.NonNull("JobType")
    status = dauphin.NonNull("JobStatus")
    repositoryOrigin = dauphin.NonNull("RepositoryOrigin")
    jobSpecificData = dauphin.Field("JobSpecificData")
    runs = dauphin.Field(dauphin.non_null_list("PipelineRun"), limit=dauphin.Int())
    runsCount = dauphin.NonNull(dauphin.Int)
    ticks = dauphin.Field(
        dauphin.non_null_list("JobTick"),
        dayRange=dauphin.Int(),
        dayOffset=dauphin.Int(),
        limit=dauphin.Int(),
    )
    nextTick = dauphin.Field("FutureJobTick")
    runningCount = dauphin.NonNull(dauphin.Int)  # remove with cron scheduler

    def __init__(self, job_state):
        self._job_state = check.inst_param(job_state, "job_state", JobState)
        super(DauphinJobState, self).__init__(
            id=job_state.job_origin_id,
            name=job_state.name,
            jobType=job_state.job_type,
            status=job_state.status,
        )

    def resolve_repositoryOrigin(self, graphene_info):
        origin = self._job_state.origin.external_repository_origin
        return graphene_info.schema.type_named("RepositoryOrigin")(origin)

    def resolve_jobSpecificData(self, graphene_info):
        if not self._job_state.job_specific_data:
            return None

        if self._job_state.job_type == JobType.SENSOR:
            return graphene_info.schema.type_named("SensorJobData")(
                self._job_state.job_specific_data
            )

        if self._job_state.job_type == JobType.SCHEDULE:
            return graphene_info.schema.type_named("ScheduleJobData")(
                self._job_state.job_specific_data
            )

        return None

    def resolve_runs(self, graphene_info, **kwargs):
        if self._job_state.job_type == JobType.SENSOR:
            filters = PipelineRunsFilter.for_sensor(self._job_state)
        else:
            filters = PipelineRunsFilter.for_schedule(self._job_state)
        return [
            graphene_info.schema.type_named("PipelineRun")(r)
            for r in graphene_info.context.instance.get_runs(
                filters=filters, limit=kwargs.get("limit"),
            )
        ]

    def resolve_runsCount(self, graphene_info):
        if self._job_state.job_type == JobType.SENSOR:
            filters = PipelineRunsFilter.for_sensor(self._job_state)
        else:
            filters = PipelineRunsFilter.for_schedule(self._job_state)
        return graphene_info.context.instance.get_runs_count(filters=filters)

    def resolve_ticks(self, graphene_info, dayRange=None, dayOffset=None, limit=None):
        before = pendulum.now("UTC").subtract(days=dayOffset).timestamp() if dayOffset else None
        after = (
            pendulum.now("UTC").subtract(days=dayRange + (dayOffset or 0)).timestamp()
            if dayRange
            else None
        )
        return [
            graphene_info.schema.type_named("JobTick")(graphene_info, tick)
            for tick in graphene_info.context.instance.get_job_ticks(
                self._job_state.job_origin_id, before=before, after=after, limit=limit
            )
        ]

    def resolve_nextTick(self, graphene_info):
        # sensor
        if self._job_state.job_type == JobType.SENSOR:
            return get_sensor_next_tick(graphene_info, self._job_state)
        else:
            return get_schedule_next_tick(graphene_info, self._job_state)

    def resolve_runningCount(self, graphene_info):
        if self._job_state.job_type == JobType.SENSOR:
            return 1 if self._job_state.status == JobStatus.RUNNING else 0
        else:
            return graphene_info.context.instance.running_schedule_count(
                self._job_state.job_origin_id
            )


class DauphinJobSpecificData(dauphin.Union):
    class Meta:
        name = "JobSpecificData"
        types = ("SensorJobData", "ScheduleJobData")


class DauphinSensorJobData(dauphin.ObjectType):
    class Meta:
        name = "SensorJobData"

    lastTickTimestamp = dauphin.Float()
    lastRunKey = dauphin.String()

    def __init__(self, job_specific_data):
        check.inst_param(job_specific_data, "job_specific_data", SensorJobData)
        super(DauphinSensorJobData, self).__init__(
            lastTickTimestamp=job_specific_data.last_tick_timestamp,
            lastRunKey=job_specific_data.last_run_key,
        )


class DauphinScheduleJobData(dauphin.ObjectType):
    class Meta:
        name = "ScheduleJobData"

    cronSchedule = dauphin.NonNull(dauphin.String)
    startTimestamp = dauphin.Float()

    def __init__(self, job_specific_data):
        check.inst_param(job_specific_data, "job_specific_data", ScheduleJobData)
        super(DauphinScheduleJobData, self).__init__(
            cronSchedule=job_specific_data.cron_schedule,
            startTimestamp=job_specific_data.start_timestamp,
        )


class DauphinJobStateOrError(dauphin.Union):
    class Meta:
        name = "JobStateOrError"
        types = ("JobState", "PythonError")


class DauphinJobStatesOrError(dauphin.Union):
    class Meta:
        name = "JobStatesOrError"
        types = ("JobStates", "PythonError")


class DauphinJobStates(dauphin.ObjectType):
    class Meta:
        name = "JobStates"

    results = dauphin.non_null_list("JobState")


DauphinJobType = dauphin.Enum.from_enum(JobType)
DauphinJobStatus = dauphin.Enum.from_enum(JobStatus)
DauphinJobTickStatus = dauphin.Enum.from_enum(JobTickStatus)



from dagster.core.scheduler.job import JobTickStatus


def tick_specific_data_from_dagster_tick(graphene_info, tick):
    if tick.status == JobTickStatus.SUCCESS:
        if tick.run_ids and graphene_info.context.instance.has_run(tick.run_ids[0]):
            return graphene_info.schema.type_named("ScheduleTickSuccessData")(
                run=graphene_info.schema.type_named("PipelineRun")(
                    graphene_info.context.instance.get_run_by_id(tick.run_ids[0])
                )
            )
        return graphene_info.schema.type_named("ScheduleTickSuccessData")(run=None)
    elif tick.status == JobTickStatus.FAILURE:
        error = tick.error
        return graphene_info.schema.type_named("ScheduleTickFailureData")(error=error)



from collections import defaultdict

from dagster import check
from dagster.core.events import DagsterEventType
from dagster.core.execution.plan.handle import StepHandle, UnresolvedStepHandle
from dagster.core.execution.plan.step import ResolvedFromDynamicStepHandle
from dagster.core.host_representation import ExternalExecutionPlan, ExternalPipeline
from dagster.core.instance import DagsterInstance
from dagster.core.storage.tags import RESUME_RETRY_TAG
from graphql.execution.base import ResolveInfo

from .external import get_external_execution_plan_or_raise
from .utils import ExecutionParams


def _update_tracking_dict(tracking, handle):
    if isinstance(handle, ResolvedFromDynamicStepHandle):
        tracking[handle.unresolved_form.to_key()].add(handle)
    else:
        tracking[handle.to_key()].add(handle)


def _in_tracking_dict(handle, tracking):
    if isinstance(handle, ResolvedFromDynamicStepHandle):
        unresolved_key = handle.unresolved_form.to_key()
        if unresolved_key in tracking:
            return handle in tracking[unresolved_key]
        else:
            return False
    else:
        return handle.to_key() in tracking


def get_retry_steps_from_execution_plan(instance, execution_plan, parent_run_id):
    check.inst_param(instance, "instance", DagsterInstance)
    check.inst_param(execution_plan, "execution_plan", ExternalExecutionPlan)
    check.opt_str_param(parent_run_id, "parent_run_id")

    if not parent_run_id:
        return execution_plan.step_keys_in_plan

    parent_run = instance.get_run_by_id(parent_run_id)
    parent_run_logs = instance.all_logs(parent_run_id)

    # keep track of steps with dicts that point:
    # * step_key -> set(step_handle) in the normal case
    # * unresolved_step_key -> set(resolved_step_handle, ...) for dynamic outputs
    all_steps_in_parent_run_logs = defaultdict(set)
    failed_steps_in_parent_run_logs = defaultdict(set)
    successful_steps_in_parent_run_logs = defaultdict(set)
    interrupted_steps_in_parent_run_logs = defaultdict(set)
    skipped_steps_in_parent_run_logs = defaultdict(set)

    for record in parent_run_logs:
        if record.dagster_event and record.dagster_event.step_handle:
            step_handle = record.dagster_event.step_handle
            _update_tracking_dict(all_steps_in_parent_run_logs, step_handle)

            if record.dagster_event_type == DagsterEventType.STEP_FAILURE:
                _update_tracking_dict(failed_steps_in_parent_run_logs, step_handle)

            if record.dagster_event_type == DagsterEventType.STEP_SUCCESS:
                _update_tracking_dict(successful_steps_in_parent_run_logs, step_handle)

            if record.dagster_event_type == DagsterEventType.STEP_SKIPPED:
                _update_tracking_dict(skipped_steps_in_parent_run_logs, step_handle)

    for step_set in all_steps_in_parent_run_logs.values():
        for step_handle in step_set:
            if (
                not _in_tracking_dict(step_handle, failed_steps_in_parent_run_logs)
                and not _in_tracking_dict(step_handle, successful_steps_in_parent_run_logs)
                and not _in_tracking_dict(step_handle, skipped_steps_in_parent_run_logs)
            ):
                _update_tracking_dict(interrupted_steps_in_parent_run_logs, step_handle)

    to_retry = defaultdict(set)

    execution_deps = execution_plan.execution_deps()
    for step_snap in execution_plan.topological_steps():
        step_key = step_snap.key
        step_handle = StepHandle.parse_from_key(step_snap.key)

        if parent_run.step_keys_to_execute and step_snap.key not in parent_run.step_keys_to_execute:
            continue

        if step_snap.key in failed_steps_in_parent_run_logs:
            to_retry[step_key].update(failed_steps_in_parent_run_logs[step_key])

        # Interrupted steps can occur when graceful cleanup from a step failure fails to run,
        # and a step failure event is not generated
        if step_key in interrupted_steps_in_parent_run_logs:
            to_retry[step_key].update(interrupted_steps_in_parent_run_logs[step_key])

        # Missing steps did not execute, e.g. when a run was terminated
        if step_key not in all_steps_in_parent_run_logs:
            to_retry[step_key].add(step_handle)

        step_dep_keys = execution_deps[step_key]
        retrying_dep_keys = step_dep_keys.intersection(to_retry.keys())

        # this step is downstream of a step we are about to retry
        if retrying_dep_keys:
            for retrying_key in retrying_dep_keys:
                # If this step and its ancestor are both downstream of a dynamic output,
                # add resolved instances of this step for the retrying mapping keys
                if isinstance(step_handle, UnresolvedStepHandle) and all(
                    map(
                        lambda handle: isinstance(handle, ResolvedFromDynamicStepHandle),
                        to_retry[retrying_key],
                    )
                ):
                    for resolved_handle in to_retry[retrying_key]:
                        to_retry[step_key].add(step_handle.resolve(resolved_handle.mapping_key))

                else:
                    to_retry[step_key].add(step_handle)

    return [step_handle.to_key() for step_set in to_retry.values() for step_handle in step_set]


def compute_step_keys_to_execute(graphene_info, external_pipeline, execution_params):
    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(external_pipeline, "external_pipeline", ExternalPipeline)
    check.inst_param(execution_params, "execution_params", ExecutionParams)

    instance = graphene_info.context.instance

    if not execution_params.step_keys and is_resume_retry(execution_params):
        # Get step keys from parent_run_id if it's a resume/retry
        external_execution_plan = get_external_execution_plan_or_raise(
            graphene_info=graphene_info,
            external_pipeline=external_pipeline,
            mode=execution_params.mode,
            run_config=execution_params.run_config,
            step_keys_to_execute=None,
        )
        return get_retry_steps_from_execution_plan(
            instance, external_execution_plan, execution_params.execution_metadata.parent_run_id
        )
    else:
        return execution_params.step_keys


def is_resume_retry(execution_params):
    check.inst_param(execution_params, "execution_params", ExecutionParams)
    return execution_params.execution_metadata.tags.get(RESUME_RETRY_TAG) == "true"



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots["TestMaterializations.test_materializations[in_memory_instance_in_process_env] 1"] = [
    "PipelineStartingEvent",
    "PipelineStartEvent",
    "ExecutionStepStartEvent",
    "StepMaterializationEvent",
    "ExecutionStepOutputEvent",
    "HandledOutputEvent",
    "ExecutionStepSuccessEvent",
    "PipelineSuccessEvent",
]

snapshots[
    "TestMaterializations.test_materializations[sqlite_with_default_run_launcher_deployed_grpc_env] 1"
] = [
    "PipelineStartingEvent",
    "PipelineStartEvent",
    "ExecutionStepStartEvent",
    "StepMaterializationEvent",
    "ExecutionStepOutputEvent",
    "HandledOutputEvent",
    "ExecutionStepSuccessEvent",
    "PipelineSuccessEvent",
]

snapshots[
    "TestMaterializations.test_materializations[sqlite_with_default_run_launcher_managed_grpc_env] 1"
] = [
    "PipelineStartingEvent",
    "PipelineStartEvent",
    "ExecutionStepStartEvent",
    "StepMaterializationEvent",
    "ExecutionStepOutputEvent",
    "HandledOutputEvent",
    "ExecutionStepSuccessEvent",
    "PipelineSuccessEvent",
]

snapshots[
    "TestMaterializations.test_materializations[sqlite_with_sync_run_launcher_in_process_env] 1"
] = [
    "PipelineStartingEvent",
    "PipelineStartEvent",
    "ExecutionStepStartEvent",
    "StepMaterializationEvent",
    "ExecutionStepOutputEvent",
    "HandledOutputEvent",
    "ExecutionStepSuccessEvent",
    "PipelineSuccessEvent",
]



from dagster import check
from dagster.core.host_representation import ExternalSensor, SensorSelector
from dagster.core.scheduler.job import JobState
from dagster_graphql import dauphin
from dagster_graphql.implementation.fetch_sensors import (
    get_sensor_next_tick,
    start_sensor,
    stop_sensor,
)
from dagster_graphql.schema.errors import (
    DauphinPythonError,
    DauphinRepositoryNotFoundError,
    DauphinSensorNotFoundError,
)


class DauphinSensor(dauphin.ObjectType):
    class Meta:
        name = "Sensor"

    id = dauphin.NonNull(dauphin.ID)
    jobOriginId = dauphin.NonNull(dauphin.String)
    name = dauphin.NonNull(dauphin.String)
    pipelineName = dauphin.NonNull(dauphin.String)
    solidSelection = dauphin.List(dauphin.String)
    mode = dauphin.NonNull(dauphin.String)
    sensorState = dauphin.NonNull("JobState")
    nextTick = dauphin.Field("FutureJobTick")

    def resolve_id(self, _):
        return "%s:%s" % (self.name, self.pipelineName)

    def __init__(self, graphene_info, external_sensor):
        self._external_sensor = check.inst_param(external_sensor, "external_sensor", ExternalSensor)
        self._sensor_state = graphene_info.context.instance.get_job_state(
            self._external_sensor.get_external_origin_id()
        )

        if not self._sensor_state:
            # Also include a SensorState for a stopped sensor that may not
            # have a stored database row yet
            self._sensor_state = self._external_sensor.get_default_job_state()

        super(DauphinSensor, self).__init__(
            name=external_sensor.name,
            jobOriginId=external_sensor.get_external_origin_id(),
            pipelineName=external_sensor.pipeline_name,
            solidSelection=external_sensor.solid_selection,
            mode=external_sensor.mode,
        )

    def resolve_sensorState(self, graphene_info):
        return graphene_info.schema.type_named("JobState")(self._sensor_state)

    def resolve_nextTick(self, graphene_info):
        return get_sensor_next_tick(graphene_info, self._sensor_state)


class DauphinSensorOrError(dauphin.Union):
    class Meta:
        name = "SensorOrError"
        types = (
            "Sensor",
            DauphinSensorNotFoundError,
            DauphinPythonError,
        )


class DauphinSensors(dauphin.ObjectType):
    class Meta:
        name = "Sensors"

    results = dauphin.non_null_list("Sensor")


class DauphinSensorsOrError(dauphin.Union):
    class Meta:
        name = "SensorsOrError"
        types = (DauphinSensors, DauphinRepositoryNotFoundError, DauphinPythonError)


class DauphinStartSensorMutation(dauphin.Mutation):
    class Meta:
        name = "StartSensorMutation"

    class Arguments:
        sensor_selector = dauphin.NonNull("SensorSelector")

    Output = dauphin.NonNull("SensorOrError")

    def mutate(self, graphene_info, sensor_selector):
        return start_sensor(graphene_info, SensorSelector.from_graphql_input(sensor_selector))


class DauphinStopSensorMutation(dauphin.Mutation):
    class Meta:
        name = "StopSensorMutation"

    class Arguments:
        job_origin_id = dauphin.NonNull(dauphin.String)

    Output = dauphin.NonNull("StopSensorMutationResultOrError")

    def mutate(self, graphene_info, job_origin_id):
        return stop_sensor(graphene_info, job_origin_id)


class DauphinStopSensorMutationResult(dauphin.ObjectType):
    class Meta:
        name = "StopSensorMutationResult"

    jobState = dauphin.Field("JobState")

    def __init__(self, job_state):
        self._job_state = check.inst_param(job_state, "job_state", JobState)

    def resolve_jobState(self, graphene_info):
        if not self._job_state:
            return None

        return graphene_info.schema.type_named("JobState")(job_state=self._job_state)


class DauphinStopSensorMutationResultOrError(dauphin.Union):
    class Meta:
        name = "StopSensorMutationResultOrError"
        types = ("StopSensorMutationResult", "PythonError")



import pendulum
from dagster import check
from dagster.core.host_representation import ExternalSchedule
from dagster.seven import get_current_datetime_in_utc, get_timestamp_from_utc_datetime
from dagster_graphql import dauphin
from dagster_graphql.schema.errors import (
    DauphinPythonError,
    DauphinRepositoryNotFoundError,
    DauphinScheduleNotFoundError,
)


class DauphinScheduleOrError(dauphin.Union):
    class Meta:
        name = "ScheduleOrError"
        types = ("Schedule", DauphinScheduleNotFoundError, DauphinPythonError)


class DauphinSchedules(dauphin.ObjectType):
    class Meta:
        name = "Schedules"

    results = dauphin.non_null_list("Schedule")


class DauphinSchedulesOrError(dauphin.Union):
    class Meta:
        name = "SchedulesOrError"
        types = (DauphinSchedules, DauphinRepositoryNotFoundError, DauphinPythonError)


class DauphinSchedule(dauphin.ObjectType):
    class Meta:
        name = "Schedule"

    id = dauphin.NonNull(dauphin.ID)
    name = dauphin.NonNull(dauphin.String)
    cron_schedule = dauphin.NonNull(dauphin.String)
    pipeline_name = dauphin.NonNull(dauphin.String)
    solid_selection = dauphin.List(dauphin.String)
    mode = dauphin.NonNull(dauphin.String)
    execution_timezone = dauphin.Field(dauphin.String)
    scheduleState = dauphin.NonNull("JobState")
    partition_set = dauphin.Field("PartitionSet")

    futureTicks = dauphin.NonNull("FutureJobTicks", cursor=dauphin.Float(), limit=dauphin.Int())
    futureTick = dauphin.NonNull("FutureJobTick", tick_timestamp=dauphin.NonNull(dauphin.Int))

    def resolve_id(self, _):
        return "%s:%s" % (self.name, self.pipeline_name)

    def resolve_partition_set(self, graphene_info):
        if self._external_schedule.partition_set_name is None:
            return None

        repository = graphene_info.context.get_repository_location(
            self._external_schedule.handle.location_name
        ).get_repository(self._external_schedule.handle.repository_name)
        external_partition_set = repository.get_external_partition_set(
            self._external_schedule.partition_set_name
        )

        return graphene_info.schema.type_named("PartitionSet")(
            external_repository_handle=repository.handle,
            external_partition_set=external_partition_set,
        )

    def resolve_futureTicks(self, graphene_info, **kwargs):
        cursor = kwargs.get(
            "cursor", get_timestamp_from_utc_datetime(get_current_datetime_in_utc())
        )
        limit = kwargs.get("limit", 10)

        tick_times = []
        time_iter = self._external_schedule.execution_time_iterator(cursor)

        for _ in range(limit):
            tick_times.append(next(time_iter).timestamp())

        future_ticks = [
            graphene_info.schema.type_named("FutureJobTick")(self._schedule_state, tick_time)
            for tick_time in tick_times
        ]

        return graphene_info.schema.type_named("FutureJobTicks")(
            results=future_ticks, cursor=tick_times[-1] + 1
        )

    def resolve_futureTick(self, graphene_info, tick_timestamp):
        return graphene_info.schema.type_named("FutureJobTick")(
            self._schedule_state, float(tick_timestamp)
        )

    def __init__(self, graphene_info, external_schedule):
        self._external_schedule = check.inst_param(
            external_schedule, "external_schedule", ExternalSchedule
        )
        self._schedule_state = graphene_info.context.instance.get_job_state(
            self._external_schedule.get_external_origin_id()
        )

        if not self._schedule_state:
            # Also include a ScheduleState for a stopped schedule that may not
            # have a stored database row yet
            self._schedule_state = self._external_schedule.get_default_job_state(
                graphene_info.context.instance
            )

        super(DauphinSchedule, self).__init__(
            name=external_schedule.name,
            cron_schedule=external_schedule.cron_schedule,
            pipeline_name=external_schedule.pipeline_name,
            solid_selection=external_schedule.solid_selection,
            mode=external_schedule.mode,
            scheduleState=graphene_info.schema.type_named("JobState")(self._schedule_state),
            execution_timezone=(
                self._external_schedule.execution_timezone
                if self._external_schedule.execution_timezone
                else pendulum.now().timezone.name
            ),
        )
